@@ -4,6 +4,9 @@ import os
 import subprocess
 from getpass import getuser
 
+# import lshell specifics
+from lshell import utils
+
 TOPDIR = os.path.abspath(os.path.join(os.path.dirname(__file__), '..'))
 
 
@@ -21,15 +24,15 @@ class TestFunctions(unittest.TestCase):
     def tearDown(self):
         self.child.close()
 
-    def test_01(self):
-        """ 01 - test lshell welcome message """
+    def test_01_welcome_message(self):
+        """ F01 | lshell welcome message """
         expected = "You are in a limited shell.\r\nType '?' or 'help' to get" \
             " the list of allowed commands\r\n"
-        result = self.child.before
+        result = self.child.before.decode('utf8')
         self.assertEqual(expected, result)
 
-    def test_02(self):
-        """ 02 - get the output of ls """
+    def test_02_builtin_ls_command(self):
+        """ F02 | built-in ls command """
         p = subprocess.Popen("ls ~",
                              shell=True,
                              stdin=subprocess.PIPE,
@@ -38,38 +41,38 @@ class TestFunctions(unittest.TestCase):
         expected = cout.read(-1)
         self.child.sendline('ls')
         self.child.expect('%s:~\$' % self.user)
-        output = self.child.before.split('ls\r', 1)[1]
+        output = self.child.before.decode('utf8').split('ls\r', 1)[1]
         self.assertEqual(len(expected.strip().split()),
                          len(output.strip().split()))
 
-    def test_03(self):
-        """ 03 - echo number """
+    def test_03_external_echo_command_num(self):
+        """ F03 | external echo number """
         expected = "32"
         self.child.sendline('echo 32')
         self.child.expect("%s:~\$" % self.user)
-        result = self.child.before.split()[2]
+        result = self.child.before.decode('utf8').split()[2]
         self.assertEqual(expected, result)
 
-    def test_04(self):
-        """ 04 - echo anything """
+    def test_04_external_echo_command_string(self):
+        """ F04 | external echo random string """
         expected = "bla blabla  32 blibli! plop."
         self.child.sendline('echo "%s"' % expected)
         self.child.expect('%s:~\$' % self.user)
-        result = self.child.before.split('\n', 1)[1].strip()
+        result = self.child.before.decode('utf8').split('\n', 1)[1].strip()
         self.assertEqual(expected, result)
 
-    def test_05(self):
-        """ 05 - echo $(uptime) """
+    def test_05_external_echo_forbidden_syntax(self):
+        """ F05 | echo forbidden syntax $(bleh) """
         expected = "*** forbidden syntax -> \"echo $(uptime)\"\r\n*** You " \
             "have 1 warning(s) left, before getting kicked out.\r\nThis " \
             "incident has been reported.\r\n"
         self.child.sendline('echo $(uptime)')
         self.child.expect('%s:~\$' % self.user)
-        result = self.child.before.split('\n', 1)[1]
+        result = self.child.before.decode('utf8').split('\n', 1)[1]
         self.assertEqual(expected, result)
 
-    def test_06_0(self):
-        """ 06.0 - change directory """
+    def test_06_builtin_cd_change_dir(self):
+        """ F06 | built-in cd - change directory """
         expected = ""
         home = os.path.expanduser('~')
         dirpath = None
@@ -82,81 +85,81 @@ class TestFunctions(unittest.TestCase):
             self.child.expect('%s:~/%s\$' % (self.user, path))
             self.child.sendline('cd ..')
             self.child.expect('%s:~\$' % self.user)
-            result = self.child.before.split('\n', 1)[1]
+            result = self.child.before.decode('utf8').split('\n', 1)[1]
             self.assertEqual(expected, result)
 
-    def test_06_1(self):
-        """ 06.1 - tilda bug """
+    def test_07_builtin_cd_tilda(self):
+        """ F07 | built-in cd - tilda bug """
         expected = "*** forbidden path -> \"/etc/passwd\"\r\n*** You have" \
             " 1 warning(s) left, before getting kicked out.\r\nThis " \
             "incident has been reported.\r\n"
         self.child.sendline('ls ~/../../etc/passwd')
         self.child.expect("%s:~\$" % self.user)
-        result = self.child.before.split('\n', 1)[1]
+        result = self.child.before.decode('utf8').split('\n', 1)[1]
         self.assertEqual(expected, result)
 
-    def test_07(self):
-        """ 07 - quotes in cd "/" """
+    def test_08_builtin_cd_quotes(self):
+        """ F08 | built-in - quotes in cd "/" """
         expected = "*** forbidden path -> \"/\"\r\n*** You have" \
             " 1 warning(s) left, before getting kicked out.\r\nThis " \
             "incident has been reported.\r\n"
         self.child.sendline('ls -ld "/"')
         self.child.expect('%s:~\$' % self.user)
-        result = self.child.before.split('\n', 1)[1]
+        result = self.child.before.decode('utf8').split('\n', 1)[1]
         self.assertEqual(expected, result)
 
-    def test_08(self):
-        """ 08 - ls ~root """
+    def test_09_external_forbidden_path(self):
+        """ F09 | external command forbidden path - ls /root """
         expected = "*** forbidden path -> \"/root/\"\r\n*** You have" \
             " 1 warning(s) left, before getting kicked out.\r\nThis " \
             "incident has been reported.\r\n"
         self.child.sendline('ls ~root')
         self.child.expect('%s:~\$' % self.user)
-        result = self.child.before.split('\n', 1)[1]
+        result = self.child.before.decode('utf8').split('\n', 1)[1]
         self.assertEqual(expected, result)
 
-    def test_09(self):
-        """ 09 - cd ~root """
+    def test_10_builtin_cd_forbidden_path(self):
+        """ F10 | built-in command forbidden path - cd ~root """
         expected = "*** forbidden path -> \"/root/\"\r\n*** You have" \
             " 1 warning(s) left, before getting kicked out.\r\nThis " \
             "incident has been reported.\r\n"
         self.child.sendline('cd ~root')
         self.child.expect('%s:~\$' % self.user)
-        result = self.child.before.split('\n', 1)[1]
+        result = self.child.before.decode('utf8').split('\n', 1)[1]
         self.assertEqual(expected, result)
 
-    def test_10(self):
-        """ 10 - empty variable 'ls "$a"/etc/passwd' """
+    def test_11_etc_passwd_1(self):
+        """ F11 | /etc/passwd: empty variable 'ls "$a"/etc/passwd' """
         expected = "*** forbidden path -> \"/etc/passwd\"\r\n*** You have" \
             " 1 warning(s) left, before getting kicked out.\r\nThis " \
             "incident has been reported.\r\n"
         self.child.sendline('ls "$a"/etc/passwd')
         self.child.expect('%s:~\$' % self.user)
-        result = self.child.before.split('\n', 1)[1]
+        result = self.child.before.decode('utf8').split('\n', 1)[1]
         self.assertEqual(expected, result)
 
-    def test_11(self):
-        """ 11 - empty variable 'ls -l .*./.*./etc/passwd' """
+    def test_12_etc_passwd_2(self):
+        """ F12 | /etc/passwd: empty variable 'ls -l .*./.*./etc/passwd' """
         expected = "*** forbidden path -> \"/etc/passwd\"\r\n*** You have" \
             " 1 warning(s) left, before getting kicked out.\r\nThis " \
             "incident has been reported.\r\n"
         self.child.sendline('ls -l .*./.*./etc/passwd')
         self.child.expect('%s:~\$' % self.user)
-        result = self.child.before.split('\n', 1)[1]
+        result = self.child.before.decode('utf8').split('\n', 1)[1]
         self.assertEqual(expected, result)
 
-    def test_12(self):
-        """ 12 - empty variable 'ls -l .?/.?/etc/passwd' """
+    def test_13_etc_passwd_3(self):
+        """ F13 | /etc/passwd: empty variable 'ls -l .?/.?/etc/passwd' """
         expected = "*** forbidden path -> \"/etc/passwd\"\r\n*** You have" \
             " 1 warning(s) left, before getting kicked out.\r\nThis " \
             "incident has been reported.\r\n"
         self.child.sendline('ls -l .?/.?/etc/passwd')
         self.child.expect('%s:~\$' % self.user)
-        result = self.child.before.split('\n', 1)[1]
+        result = self.child.before.decode('utf8').split('\n', 1)[1]
         self.assertEqual(expected, result)
 
-    def test_13(self):
-        """ 13 - completion with ~/ """
+    def test_14_path_completion_tilda(self):
+        """ F14 | path completion with ~/ """
         p = subprocess.Popen("ls -F ~/",
                              shell=True,
                              stdin=subprocess.PIPE,
@@ -165,46 +168,41 @@ class TestFunctions(unittest.TestCase):
         expected = cout.read(-1)
         self.child.sendline('cd ~/\t\t')
         self.child.expect('%s:~\$' % self.user)
-        output = self.child.before.split('\n', 1)[1]
+        output = self.child.before.decode('utf8').split('\n', 1)[1]
         self.assertEqual(len(expected.strip().split()),
                          len(output.strip().split()))
 
-#    def test_14(self):
-#        """ 14 - command over ssh """
-
-    def test_15(self):
-        """ 15 - tab to list commands """
-        expected = '\x07\r\ncd       echo     help     ll       ls       \r\n'\
-            'clear    exit     history  lpath    lsudo'
+    def test_15_cmd_completion_tab_tab(self):
+        """ F15 | command completion: tab to list commands """
+        expected = '\x07\r\ncd       echo     help     ll       ls       '\
+                   '\r\nclear    exit     history  lpath    lsudo'
         self.child.sendline('\t\t')
         self.child.expect('%s:~\$' % self.user)
-        result = self.child.before.strip()
+        result = self.child.before.decode('utf8').strip()
+
+        print(expected)
+        print(result)
         self.assertEqual(expected, result)
 
-#    def test_exit(self):
-#        expected = ''
-#        self.child.sendline('exit')
-#        self.child.expect('$')
-#        result = self.child.before
-#        self.assertEqual(expected, result)
-
-    def test_16_exitcode_with_separator(self):
-        """ 16 - test exit codes with separator """
+    def test_16_exitcode_with_separator_external_cmd(self):
+        """ F16 | external command exit codes with separator """
         self.child = pexpect.spawn('%s/bin/lshell '
-                                   '--config %s/etc/lshell.conf --forbidden "[]"'
+                                   '--config %s/etc/lshell.conf '
+                                   '--forbidden "[]"'
                                    % (TOPDIR, TOPDIR))
         self.child.expect('%s:~\$' % self.user)
 
         expected = "2"
         self.child.sendline('ls nRVmmn8RGypVneYIp8HxyVAvaEaD55; echo $?')
         self.child.expect('%s:~\$' % self.user)
-        result = self.child.before.split('\n')[2].strip()
+        result = self.child.before.decode('utf8').split('\n')[2].strip()
         self.assertEqual(expected, result)
 
-    def test_17_exitcode_without_separator(self):
-        """ 17 - test exit codes without separator """
+    def test_17_exitcode_without_separator_external_cmd(self):
+        """ F17 | external command exit codes without separator """
         self.child = pexpect.spawn('%s/bin/lshell '
-                                   '--config %s/etc/lshell.conf --forbidden "[]"'
+                                   '--config %s/etc/lshell.conf '
+                                   '--forbidden "[]"'
                                    % (TOPDIR, TOPDIR))
         self.child.expect('%s:~\$' % self.user)
 
@@ -213,13 +211,66 @@ class TestFunctions(unittest.TestCase):
         self.child.expect('%s:~\$' % self.user)
         self.child.sendline('echo $?')
         self.child.expect('%s:~\$' % self.user)
-        result = self.child.before.split('\n')[1].strip()
+        result = self.child.before.decode('utf8').split('\n')[1].strip()
         self.assertEqual(expected, result)
 
-    def test_18_allow_slash(self):
-        """ 18 - user should able to allow / access minus some directory (e.g. /var) """
+    def test_18_cd_exitcode_with_separator_internal_cmd(self):
+        """ F18 | built-in command exit codes with separator """
         self.child = pexpect.spawn('%s/bin/lshell '
-                                   '--config %s/etc/lshell.conf --path "[\'/\'] - [\'/var\']"'
+                                   '--config %s/etc/lshell.conf '
+                                   '--forbidden "[]"'
+                                   % (TOPDIR, TOPDIR))
+        self.child.expect('%s:~\$' % self.user)
+
+        expected = "2"
+        self.child.sendline('cd nRVmmn8RGypVneYIp8HxyVAvaEaD55; echo $?')
+        self.child.expect('%s:~\$' % self.user)
+        self.child.sendline('echo $?')
+        self.child.expect('%s:~\$' % self.user)
+        result = self.child.before.decode('utf8').split('\n')[1].strip()
+        self.assertEqual(expected, result)
+
+    def test_19_cd_exitcode_without_separator_external_cmd(self):
+        """ F19 | built-in exit codes without separator """
+        self.child = pexpect.spawn('%s/bin/lshell '
+                                   '--config %s/etc/lshell.conf '
+                                   '--forbidden "[]"'
+                                   % (TOPDIR, TOPDIR))
+        self.child.expect('%s:~\$' % self.user)
+
+        expected = "2"
+        self.child.sendline('cd nRVmmn8RGypVneYIp8HxyVAvaEaD55')
+        self.child.expect('%s:~\$' % self.user)
+        self.child.sendline('echo $?')
+        self.child.expect('%s:~\$' % self.user)
+        result = self.child.before.decode('utf8').split('\n')[1].strip()
+        self.assertEqual(expected, result)
+
+    def test_20_cd_with_cmd_unknwon_dir(self):
+        """ F20 | test built-in cd with command when dir does not exist
+            Should be returning error, not executing cmd
+        """
+        self.child = pexpect.spawn('%s/bin/lshell '
+                                   '--config %s/etc/lshell.conf '
+                                   '--forbidden "[]"'
+                                   % (TOPDIR, TOPDIR))
+        self.child.expect('%s:~\$' % self.user)
+
+        expected = 'lshell: nRVmmn8RGypVneYIp8HxyVAvaEaD55: No such file or '\
+                   'directory'
+
+        self.child.sendline('cd nRVmmn8RGypVneYIp8HxyVAvaEaD55; echo $?')
+        self.child.expect('%s:~\$' % self.user)
+        result = self.child.before.decode('utf8').split('\n')[1].strip()
+        self.assertEqual(expected, result)
+
+    def test_21_allow_slash(self):
+        """ F21 | user should able to allow / access minus some directory
+            (e.g. /var)
+        """
+        self.child = pexpect.spawn('%s/bin/lshell '
+                                   '--config %s/etc/lshell.conf '
+                                   '--path "[\'/\'] - [\'/var\']"'
                                    % (TOPDIR, TOPDIR))
         self.child.expect('%s:~\$' % self.user)
 
@@ -228,13 +279,14 @@ class TestFunctions(unittest.TestCase):
         self.child.expect('%s:/\$' % self.user)
         self.child.sendline('cd var')
         self.child.expect('%s:/\$' % self.user)
-        result = self.child.before.split('\n')[1].strip()
+        result = self.child.before.decode('utf8').split('\n')[1].strip()
         self.assertEqual(expected, result)
 
-    def test_19_expand_env_variables(self):
-        """ 19 - test expanding of environment variables """
+    def test_22_expand_env_variables(self):
+        """ F22 | expanding of environment variables """
         self.child = pexpect.spawn('%s/bin/lshell '
-                                   '--config %s/etc/lshell.conf --allowed "+ [\'export\']"'
+                                   '--config %s/etc/lshell.conf '
+                                   '--allowed "+ [\'export\']"'
                                    % (TOPDIR, TOPDIR))
         self.child.expect('%s:~\$' % self.user)
 
@@ -243,32 +295,31 @@ class TestFunctions(unittest.TestCase):
         self.child.expect('%s:~\$' % self.user)
         self.child.sendline('echo $HOME/$A')
         self.child.expect('%s:~\$' % self.user)
-        result = self.child.before.split('\n')[1].strip()
+        result = self.child.before.decode('utf8').split('\n')[1].strip()
         self.assertEqual(expected, result)
 
-    def test_20_expand_env_variables_cd(self):
-        """ 20 - test expanding of environment variables when using cd """
+    def test_23_expand_env_variables_cd(self):
+        """ F23 | expanding of environment variables when using cd """
         self.child = pexpect.spawn('%s/bin/lshell '
-                                   '--config %s/etc/lshell.conf --allowed "+ [\'export\']"'
+                                   '--config %s/etc/lshell.conf '
+                                   '--allowed "+ [\'export\']"'
                                    % (TOPDIR, TOPDIR))
         self.child.expect('%s:~\$' % self.user)
 
-        import random
-        import string
+        random = utils.random_string(32)
 
-        random = ''.join([random.choice(string.ascii_letters + string.digits) for n in xrange(32)])
-
-        expected = "lshell: %s/random_%s: No such file or directory" \
-                                            % (os.path.expanduser('~'),random)
+        expected = 'lshell: %s/random_%s: No such file or directory' % (
+            os.path.expanduser('~'), random)
         self.child.sendline('export A=random_%s' % random)
         self.child.expect('%s:~\$' % self.user)
         self.child.sendline('cd $HOME/$A')
         self.child.expect('%s:~\$' % self.user)
-        result = self.child.before.split('\n')[1].strip()
+        result = self.child.before.decode('utf8').split('\n')[1].strip()
         self.assertEqual(expected, result)
 
-    def test_21_cd_and_command(self):
-        """ 07 - cd && command should not be interpreted by internal function """
+    def test_24_cd_and_command(self):
+        """ F24 | cd && command should not be interpreted by internal function
+        """
         self.child = pexpect.spawn('%s/bin/lshell '
                                    '--config %s/etc/lshell.conf'
                                    % (TOPDIR, TOPDIR))
@@ -277,22 +328,25 @@ class TestFunctions(unittest.TestCase):
         expected = "OK"
         self.child.sendline('cd ~ && echo "OK"')
         self.child.expect('%s:~\$' % self.user)
-        result = self.child.before.split('\n')[1].strip()
+        result = self.child.before.decode('utf8').split('\n')[1].strip()
         self.assertEqual(expected, result)
 
-    def test_22_KeyboardInterrupt(self):
-        """ 07 - test cat(1) with KeyboardInterrupt, should not exit """
+    def test_25_KeyboardInterrupt(self):
+        """ F25 | test cat(1) with KeyboardInterrupt, should not exit """
         self.child = pexpect.spawn('%s/bin/lshell '
-                                   '--config %s/etc/lshell.conf --allowed "+ [\'cat\']"'
+                                   '--config %s/etc/lshell.conf '
+                                   '--allowed "+ [\'cat\']"'
                                    % (TOPDIR, TOPDIR))
         self.child.expect('%s:~\$' % self.user)
 
-        expected = "^C"
+        expected = "foo"
         self.child.sendline('cat')
-        self.child.sendcontrol('c');
+        self.child.sendline(' foo ')
+        self.child.sendcontrol('c')
         self.child.expect('%s:~\$' % self.user)
-        result = self.child.before.split('\n')[1].strip()
-        self.assertEqual(expected, result)
+        result = self.child.before.decode('utf8').split('\n')[1].strip()
+        self.assertIn(expected, result)
+
 
 if __name__ == '__main__':
     unittest.main()
